import logging
import subprocess

from ci_ynh.models import Check, Package


logger = logging.getLogger(__name__)


def call_yunohost(*, action_name, args, package, check):
    logger.info('Start %r subprocess for "%s" with: %r', action_name, package, ' '.join(args))
    try:
        process = subprocess.run(
            args,
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
            timeout=15 * 60,
            check=True,
            text=True,
        )
    except subprocess.CalledProcessError as err:
        logger.exception('%s "%s" error: %s', action_name, package, err)
        check.status = Check.STATUS_FAIL
        check.output += f'\n{err.output}\n(return code: {err.returncode!r})'
        check.save()
        return check
    except subprocess.TimeoutExpired as err:
        logger.exception('%s "%s" timeout', action_name, package)
        check.status = Check.STATUS_FAIL
        check.output += f'\nTimeout ({err.timeout}sec.):\n{err.output}'
        check.save()
        return check

    returncode = process.returncode
    output = process.stdout
    if returncode != 0:
        logger.error('%s "%s" return code: %r', action_name, package, returncode)
        check.status = Check.STATUS_FAIL
        check.output += f'\n{action_name} failed:\n{output}\n(return code: {returncode!r})'
        check.save()
        return check

    check.output += f'\n{action_name} success:\n{output}\n(return code: {returncode!r})'
    check.save()

    return check


def check_package(*, package, check):
    assert isinstance(package, Package)
    assert isinstance(check, Check)

    args = (
        'sudo',
        'yunohost',
        'app',
        'install',
        package.url,
        '--label',
        f'{package.project_name}-{package.branch_name}',
        '--args',
        package.get_args(),
        '--force',  # Do not ask confirmation if the app is not safe to use
    )
    check = call_yunohost(action_name='install', args=args, package=package, check=check)
    if check.status is not Check.STATUS_RUNNING:
        # failed -> abort other steps
        return check

    args = (
        'sudo',
        'yunohost',
        'app',
        'remove',
        f'{package.project_name}-{package.branch_name}',
    )
    check = call_yunohost(action_name='remove', args=args, package=package, check=check)
    if check.status is not Check.STATUS_RUNNING:
        # failed -> abort other steps
        return check

    check.status = Check.STATUS_SUCCESS
    check.save()

    return check



import datetime
import logging
from uuid import UUID

from django.utils import timezone
from huey import crontab
from huey.contrib.djhuey import lock_task, periodic_task, task

from ci_ynh.models import Check, Package
from ci_ynh.yunohost_utils import check_package


logger = logging.getLogger(__name__)


@periodic_task(
    crontab(minute='*'),  # run every minutes
    context=True,  # include the Task instance as a keyword argument
)
@lock_task('schedule_ci_pipeline')  # no multiple invocations from running concurrently
def schedule_ci_pipeline(task):
    """
    Schedule activate packages for a CI run
    """
    # Set outdated checks to timeout:
    qs = Check.objects.filter(status=Check.STATUS_RUNNING)
    qs = qs.filter(update_dt__lt=timezone.now() - datetime.timedelta(minutes=30))
    qs.update(status=Check.STATUS_TIMEOUT)

    # Get all active and not running packages:
    qs = Package.objects.filter(ci_active=True)
    qs = qs.exclude(checks__status=Check.STATUS_RUNNING)

    # TODO: Order oldest CI run to first

    # Schedule filtered packages:
    for package in qs:
        logger.info('Schedule "%s"', package)
        ci_run(package_pk=package.pk)


@task(context=True)  # include the Task instance as a keyword argument
@lock_task('ci_run')  # no multiple invocations from running concurrently
def ci_run(*, package_pk, task):
    """
    CI run for one package
    """
    task_id = UUID(task.id)
    package = Package.objects.get(pk=package_pk)

    if package.ci_active is not True:
        # Should not happen
        logger.warning('Skip deactivated package: "%s"', package)
        return

    logger.info('CI run "%s"', package)
    check = Check.objects.create(package=package, task_id=task_id, status=Check.STATUS_RUNNING)

    check = check_package(package=package, check=check)
    logger.info('CI run "%s" done: %s', package, check)



from huey import RedisExpireHuey
from redis import ConnectionPool


pool = ConnectionPool(
    host='127.0.0.1',
    port=6379,
    max_connections=20,
)
HUEY = RedisExpireHuey(
    name='ci_ynh',  # Just the name for this task queue.
    connection_pool=pool,  # Use a connection pool to redis.
    results=True,  # Store return values of tasks.
    store_none=False,  # If a task returns None, do not save to results.
    utc=True,  # Use UTC for all times internally.
    expire_time=24 * 60 * 60,  # cleaned-up unread task results from Redis after 24h
)



import sys
from uuid import UUID

from django.core.management.base import BaseCommand

from ci_ynh.models import Check, Package
from ci_ynh.yunohost_utils import check_package


class Command(BaseCommand):
    help = 'Start a CI run in foreground'

    def add_arguments(self, parser):
        parser.add_argument('project_name')

    def handle(self, *args, **options):
        project_name = options['project_name']
        print(f'Start CI run for project: {project_name!r}...')

        package = Package.objects.filter(project_name=project_name).first()
        if package is None:
            print(f'ERROR: project name {project_name!r} does not exists!')
            print('Registered names are:')
            names = Package.objects.values_list('project_name', flat=True)
            for name in names:
                print(f' * {name!r}')
            sys.exit(-1)

        print(package)
        check = Check.objects.create(
            package=package, task_id=UUID('00000000-1111-0000-0000-000000000001'), status=Check.STATUS_RUNNING
        )

        check = check_package(package=package, check=check)
        print(f'CI run "{package}" done: {check}')



from django.contrib import admin
from django.utils.translation import gettext_lazy as _

from ci_ynh.models import Check, Package
from ci_ynh.tasks import ci_run


def force_ci_run(modeladmin, request, queryset):
    class FakeTask:
        id = '00000000-1111-0000-0000-000000000001'

    for package in queryset:
        ci_run(package_pk=package.pk, task=FakeTask)


force_ci_run.short_description = _('Force CI run for selected entries')


@admin.register(Package)
class PackageAdmin(admin.ModelAdmin):
    actions = [force_ci_run]


@admin.register(Check)
class CheckAdmin(admin.ModelAdmin):
    readonly_fields = ('task_id', 'output', 'status')



from bx_py_utils.models.timetracking import TimetrackingBaseModel
from django.db import models
from django.utils.translation import gettext_lazy as _


class Package(TimetrackingBaseModel):
    """
    Store information about what project should be checked.
    """

    project_name = models.CharField(
        max_length=128,
        verbose_name=_('Package.project_name.verbose_name'),
        help_text=_('Package.project_name.help_text'),
    )
    ci_active = models.BooleanField(
        default=True,
        verbose_name=_('Package.ci_active.verbose_name'),
        help_text=_('Package.ci_active.help_text'),
    )
    url = models.URLField(
        # e.g.:
        # https://github.com/YunoHost-Apps/django_ynh/tree/master
        # https://github.com/YunoHost-Apps/django_ynh/tree/testing
        unique=True,
        verbose_name=_('Package.url.verbose_name'),
        help_text=_('Package.url.help_text'),
    )
    branch_name = models.CharField(
        # e.g.:
        # master
        # testing
        max_length=64,
        verbose_name=_('Package.branch_name.verbose_name'),
        help_text=_('Package.branch_name.help_text'),
    )
    args = models.TextField(
        # TODO: generate dynamically
        default='domain=domain.tld\npath=/path\nadmin=johndoe',
        verbose_name=_('Package.args.verbose_name'),
        help_text=_('Package.args.help_text'),
    )

    def get_args(self):
        args = []
        for line in str(self.args).splitlines():
            line = line.strip()
            if line:
                args.append(line)
        return '&'.join(args)

    def __str__(self):
        return f'{self.project_name} - {self.branch_name}'

    class Meta:
        verbose_name = _('Package')
        verbose_name_plural = _('Packages')


class Check(TimetrackingBaseModel):
    """
    Store information about a check run.
    """

    package = models.ForeignKey(
        Package,
        on_delete=models.CASCADE,
        related_name='checks',
        verbose_name=_('Check.package.verbose_name'),
        help_text=_('Check.package.help_text'),
    )
    task_id = models.UUIDField(
        editable=False, verbose_name=_('Check.task_id.verbose_name'), help_text=_('Check.task_id.help_text')
    )
    output = models.TextField(
        editable=False, verbose_name=_('Check.output.verbose_name'), help_text=_('Check.output.help_text')
    )

    STATUS_UNKNOWN = 'unknown'
    STATUS_RUNNING = 'running'
    STATUS_SUCCESS = 'success'
    STATUS_FAIL = 'fail'
    STATUS_TIMEOUT = 'timeout'
    STATUS_CHOICES = (
        (STATUS_UNKNOWN, _('Unknown')),
        (STATUS_RUNNING, _('CI is Running')),
        (STATUS_SUCCESS, _('CI run successful')),
        (STATUS_FAIL, _('CI run failed')),
        (STATUS_TIMEOUT, _('Timeout')),
    )
    STATUS_CHOICES_DICT = dict(STATUS_CHOICES)
    STATUS_MAX_LENGTH = max(len(key) for key in STATUS_CHOICES_DICT.keys() if key is not None)

    status = models.CharField(
        choices=STATUS_CHOICES,
        max_length=STATUS_MAX_LENGTH,
        default=STATUS_UNKNOWN,
        blank=True,
        null=True,
        editable=False,
        verbose_name=_('Check.status.verbose_name'),
        help_text=_('Check.status.help_text'),
    )

    def __str__(self):
        return f'{self.package} - {self.status}'

    class Meta:
        verbose_name = _('Pipeline status')
        verbose_name_plural = _('Pipeline status')



# Generated by Django 3.1.5 on 2021-01-08 05:44

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='Package',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('create_dt', models.DateTimeField(blank=True, editable=False, help_text='ModelTimetrackingMixin.create_dt.help_text', null=True, verbose_name='ModelTimetrackingMixin.create_dt.verbose_name')),
                ('update_dt', models.DateTimeField(blank=True, editable=False, help_text='ModelTimetrackingMixin.update_dt.help_text', null=True, verbose_name='ModelTimetrackingMixin.update_dt.verbose_name')),
                ('project_name', models.CharField(help_text='Package.project_name.help_text', max_length=128, verbose_name='Package.project_name.verbose_name')),
                ('url', models.URLField(help_text='Package.url.help_text', unique=True, verbose_name='Package.url.verbose_name')),
                ('branch_name', models.CharField(help_text='Package.branch_name.help_text', max_length=64, verbose_name='Package.branch_name.verbose_name')),
                ('args', models.TextField(default='domain=domain.tld\npath=/path', help_text='Package.args.help_text', verbose_name='Package.args.verbose_name')),
            ],
            options={
                'verbose_name': 'Package',
                'verbose_name_plural': 'Packages',
            },
        ),
        migrations.CreateModel(
            name='Check',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('create_dt', models.DateTimeField(blank=True, editable=False, help_text='ModelTimetrackingMixin.create_dt.help_text', null=True, verbose_name='ModelTimetrackingMixin.create_dt.verbose_name')),
                ('update_dt', models.DateTimeField(blank=True, editable=False, help_text='ModelTimetrackingMixin.update_dt.help_text', null=True, verbose_name='ModelTimetrackingMixin.update_dt.verbose_name')),
                ('task_id', models.UUIDField(editable=False, help_text='Check.task_id.help_text', verbose_name='Check.task_id.verbose_name')),
                ('output', models.TextField(editable=False, help_text='Check.output.help_text', verbose_name='Check.output.verbose_name')),
                ('status', models.CharField(blank=True, choices=[('unknown', 'Unknown'), ('running', 'CI is Running'), ('success', 'CI run successful'), ('fail', 'CI run failed')], default='unknown', editable=False, help_text='Check.status.help_text', max_length=7, null=True, verbose_name='Check.status.verbose_name')),
                ('package', models.ForeignKey(help_text='Check.package.help_text', on_delete=django.db.models.deletion.CASCADE, related_name='checks', to='ci_ynh.package', verbose_name='Check.package.verbose_name')),
            ],
            options={
                'verbose_name': 'Pipeline status',
                'verbose_name_plural': 'Pipeline status',
            },
        ),
    ]



# Generated by Django 3.1.5 on 2021-01-08 11:04

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('ci_ynh', '0001_initial'),
    ]

    operations = [
        migrations.AddField(
            model_name='package',
            name='ci_active',
            field=models.BooleanField(default=True, help_text='Package.ci_active.help_text', verbose_name='Package.ci_active.verbose_name'),
        ),
        migrations.AlterField(
            model_name='check',
            name='status',
            field=models.CharField(blank=True, choices=[('unknown', 'Unknown'), ('running', 'CI is Running'), ('success', 'CI run successful'), ('fail', 'CI run failed'), ('timeout', 'Timeout')], default='unknown', editable=False, help_text='Check.status.help_text', max_length=7, null=True, verbose_name='Check.status.verbose_name'),
        ),
        migrations.AlterField(
            model_name='package',
            name='args',
            field=models.TextField(default='domain=domain.tld\npath=/path\nadmin=johndoe', help_text='Package.args.help_text', verbose_name='Package.args.verbose_name'),
        ),
    ]
